""" wenxin exceptions"""
import wenxin_api


class WenxinError(Exception):
    """ base wenxin error """
    def __init__(
        self,
        message=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
    ):
        super(WenxinError, self).__init__(message)

        if http_body and hasattr(http_body, "decode"):
            try:
                http_body = http_body.decode("utf-8")
            except BaseException:
                http_body = (
                    "could not decode body as utf-8."
                )

        self._message = message
        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}
        self.code = code
        self.request_id = self.headers.get("request-id", None)

    def __str__(self):
        msg = self._message or "<empty message>"
        if self.request_id is not None:
            return "Request {0}: {1}".format(self.request_id, msg)
        else:
            return msg

    @property
    def user_message(self):
        """ user message """
        return self._message

    def __repr__(self):
        return "%s(message=%r, http_status=%r, request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.http_status,
            self.request_id,
        )


class APIError(WenxinError):
    """ error returns by api server """
    pass


class TimeOutError(WenxinError):
    """ timeout error """
    pass


class NotReady(WenxinError):
    """ not ready error """
    pass


class APIConnectionError(WenxinError):
    """ api connection error """
    def __init__(
        self,
        message,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
        code=None,
        should_retry=False,
    ):
        super(APIConnectionError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.should_retry = should_retry


class InvalidRequestError(WenxinError):
    """ invalid request error """
    def __init__(
        self,
        message,
        param,
        code=None,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(InvalidRequestError, self).__init__(
            message, http_body, http_status, json_body, headers, code
        )
        self.param = param

    def __repr__(self):
        return "%s(message=%r, param=%r, code=%r, http_status=%r, " "request_id=%r)" % (
            self.__class__.__name__,
            self._message,
            self.param,
            self.code,
            self.http_status,
            self.request_id,
        )

    def __reduce__(self):
        return type(self), (
            self._message,
            self.param,
            self.code,
            self.http_body,
            self.http_status,
            self.json_body,
            self.headers,
        )

class FileError(WenxinError):
    """ file error """
    pass

class IllegalRequestArgumentError(WenxinError):
    """ illeagal request argument error """
    pass


class MissingRequestArgumentError(WenxinError):
    """ missing request argument error """
    pass


class InvalidResponseValue(WenxinError):
    """ invalid response value error """
    pass


class InternalServerError(WenxinError):
    """ internal server error """
    pass


class AuthenticationError(WenxinError):
    """ authentication error """
    pass

class AccessTokenExpiredError(WenxinError):
    """ access_token expired error """
    pass

class RateLimitError(WenxinError):
    """ qps limit error """
    pass

class ResponseDecodeError(WenxinError):
    """ response decode error """
    pass


class ServiceUnavailableError(WenxinError):
    """ service unavailable error """
    pass







import os
import sys
import logging
import logging.config
import re
from os.path import dirname
from logging.handlers import RotatingFileHandler


cur_dir = os.path.split(os.path.realpath(__file__))[0]
LOG_DIR = os.environ['LOG_DIR'] if 'LOG_DIR' in os.environ \
    and os.environ['LOG_DIR'] else os.path.join(cur_dir, "log/")
if not os.path.exists(LOG_DIR):
    os.makedirs(LOG_DIR)

LEVEL = os.environ['LOG_LEVEL'] if 'LOG_LEVEL' in os.environ \
    and os.environ['LOG_LEVEL'] else "INFO"

LOG_CONFIG = {
    'version': 1, 
    'disable_existing_loggers': False, 
    'formatters': {
        'simple': {
            'format': '%(asctime)s - %(message)s'
        },
        'standard': {
            'format': '%(asctime)s \"%(pathname)s:%(lineno)d\" [%(levelname)s]- %(message)s'
        }
    }, 
    'handlers': {
        'stream_logger': {
            'level': 'DEBUG', 
            'formatter': 'simple', 
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout'
        }, 

        'one_day_logger': {
            'level': 'DEBUG', 
            'formatter': 'standard', 
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': os.path.join(LOG_DIR, 'wenxin_api.log'), 
            'when': 'midnight', 
            'interval': 1, 
            'encoding': 'utf-8', 
        }, 
        'half_gb_logger': {
            'level': "DEBUG",
            'formatter': 'standard', 
            'class': 'logging.handlers.RotatingFileHandler', 
            'filename': os.path.join(LOG_DIR, 'wenxin_api.log'),
            'maxBytes': 1024 * 1024 * 500, 
            # https://stackoverflow.com/questions/40150821/in-the-logging-modules-rotatingfilehandler-how-to-set-the-backupcount-to-a-pra
            # clean your logs 
            # after it grows to 500G(maxBytes * backupCount)
            # RotatingFileHandler won't do rotate anymore
            'backupCount': 1000, 
            'encoding': 'utf-8', 
        }
    }, 
    'loggers': {
         'wenxin_api': {
            'handlers': ['stream_logger', 'one_day_logger', 'half_gb_logger'], 
            # FIXME 
            'level': LEVEL, 
            'propagate': True, 
         }
    }
}

logging.config.dictConfig(LOG_CONFIG)

def get_logger(name=None):
    """ get logger """
    if name is None:
        name = "wenxin_api"
    logger = logging.getLogger(name)
    return logger



# !/usr/bin/env python3
import json
import platform
import threading
import warnings
import time
from json import JSONDecodeError
from typing import Dict, Iterator, Optional, Tuple, Union
from urllib.parse import urlencode, urlsplit, urlunsplit

import requests

import wenxin_api
from wenxin_api import error, log
from wenxin_api.base_object import BaseObject
from wenxin_api.error import (APIError, 
                              APIConnectionError, 
                              InvalidResponseValue, 
                              AuthenticationError,
                              AccessTokenExpiredError,
                              ResponseDecodeError)
from wenxin_api.const import ERNIE_1P5B_MODEL, ERNIE_100B_MODEL
import wenxin_api.variable


# Has one attribute per thread, 'session'.
_thread_context = threading.local()
logger = log.get_logger()

from typing import Optional


class WenxinAPIResponse(BaseObject):
    def __init__(self, headers, request_type, **params):
        super(WenxinAPIResponse, self).__init__(**params)
        self._headers = headers
        self.type = request_type

    def __str__(self):
        return "WenxinAPIResponse {}:{}\n".format(
                        id(self),
                        json.dumps({"id": self.id, 
                                    "status": self.status,
                                    "type": self.type
                                   }, ensure_ascii=False)
        )

    def __repr__(self):
        return self.__str__()


def _requests_proxies_arg(proxy) -> Optional[Dict[str, str]]:
    if proxy is None:
        return None
    elif isinstance(proxy, str):
        return {"http": proxy, "https": proxy}
    elif isinstance(proxy, dict):
        return proxy.copy()
    else:
        raise ValueError(
            "'wenxin.proxy' must be url str or dict"
        )


def _make_session() -> requests.Session:
    s = requests.Session()
    proxies = _requests_proxies_arg(wenxin_api.proxy)
    if proxies:
        s.proxies = proxies
    s.mount(
        "https://",
        requests.adapters.HTTPAdapter(max_retries=wenxin_api.variable.MAX_CONNECTION_RETRIES),
    )
    return s


class HTTPRequestor:
    """ HTTP Requestor """ 
    def __init__(self, ak=None, sk=None, request_type=None):
        self.ak = ak
        self.sk = sk
        self.version = wenxin_api.version
        self.request_type = request_type
        self.access_token_post_url = wenxin_api.variable.ACCESS_TOKEN_URL
        
    def _get_access_token(self, ak=None, sk=None):
        if ak == None or sk == None:
            ak = wenxin_api.ak
            sk = wenxin_api.sk
        json_data = {
            "grant_type": "client_credentials",
            "client_id": ak,
            "client_secret": sk
        }
        method = "post" # hard code
        result = self._request(self.access_token_post_url, method, data=json_data)
        err_code = result.json()["code"]
        if err_code != 0:
            raise AuthenticationError(result.json()["msg"])
        logger.debug("result of access token: {}".format(result.json()))
        return result.json()["data"]

    def request(self, url, method="post", headers=None, files=None, request_id=None, **params
    ) -> Union[WenxinAPIResponse, Iterator[WenxinAPIResponse]]:
        
        if isinstance(params, dict):
            data = params
            base_model = params.get("base_model", ERNIE_100B_MODEL)
            return_raw = params.pop("return_raw", False)
        else:
            data = {}
            base_model = ERNIE_100B_MODEL

        data["cmd"] = request_id
        data["base_model"] = base_model
        data["sdk_version"] = self.version
        logger.debug("request id: {}, base_model: {}, request params: {}".format(
            request_id,
            base_model, 
            params))
        # try to use default ak/sk first
        try:
            if wenxin_api.ak != None and wenxin_api.sk != None:
                access_token = self._get_access_token(self.ak, self.sk)
                data["access_token"] =  access_token
            elif wenxin_api.access_token != None:
                data["access_token"] =  wenxin_api.access_token
            else:
                raise AuthenticationError()

            result = self._request(url, method=method, headers=headers, data=data, files=files)
            logger.debug("request body:{}".format(json.dumps({
                    "url": url,
                    "method": method,
                    "headers": headers,
                    "data": data,
                    "files": len(files) if files else None},
                    ensure_ascii=False,
                    indent=2))
            )
            logger.debug("response: code {}: {}".format(
                result.status_code,
                json.dumps(result.json(), 
                           ensure_ascii=False, 
                           indent=2)
            ))

            if return_raw:
                return result
            else:
                resp = self._resolve_response(result.content, result.status_code, result.headers)
                return resp

        except JSONDecodeError as e:
            raise APIError("get access token failed: {}\n{}".format(result, result.text))

        

    def _request(self, url, method="post", headers=None, data=None, files=None) -> requests.Response:
        if headers is None:
            headers = {"Content-Type": "application/json"}
        if not hasattr(_thread_context, "session"):
            _thread_context.session = _make_session()
        try:
            if isinstance(headers, dict) and \
               "Content-Type" in headers and \
               headers["Content-Type"] == "application/json":
                result = _thread_context.session.request(
                    method,
                    url,
                    headers=headers,
                    json=data,
                    timeout=wenxin_api.variable.TIMEOUT_SECS,
                )
            else:
                result = _thread_context.session.request(
                    method,
                    url,
                    headers=headers,
                    data=data,
                    files=files,
                    timeout=wenxin_api.variable.TIMEOUT_SECS,
                )
        except requests.exceptions.RequestException as e:
            raise error.APIConnectionError("error communicating with wenxin api") from e
        return result

    def _resolve_response(self, rbody, rcode, rheaders) -> WenxinAPIResponse:
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            data = json.loads(rbody)
        except (JSONDecodeError, UnicodeDecodeError):
            raise error.APIError(
                f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders
            )

        if rcode != 200 or data["code"] != 0:
            raise self._request_error_handling(rbody, rcode, data, rheaders)

        if len(data["data"]) == 0:
            # some method ex. delete returns null data
            resp = WenxinAPIResponse(rheaders, self.request_type)
        elif self.request_type not in data["data"]:
            # some method  ex. create only returns ${type}_id
            resp = WenxinAPIResponse(rheaders, self.request_type, **data["data"])
        elif isinstance(data["data"][self.request_type], dict):
            resp = WenxinAPIResponse(rheaders, self.request_type, **data["data"][self.request_type])
        elif isinstance(data["data"][self.request_type], list):
            resp = [
                WenxinAPIResponse(rheaders, self.request_type, **one_data) \
                    for one_data in data["data"][self.request_type]
            ]
        else:
            raise ResponseDecodeError(json.dumps(data, ensure_ascii=False, indent=2))

        return resp

    def _request_error_handling(self, rbody, rcode, resp, rheaders):

        logger.error("wenxin api error {}: {}".format(rcode, rbody))
        error_msg = resp.get("error", "")

        if rcode == 200:
            msg = "API error code {}: {}".format(resp["code"], resp["msg"])
            return error.APIError(msg)

        elif rcode == 500:
            body_code = resp.get("code", None)
            if body_code == 18:
                return error.RateLimitError()
            elif body_code == 100 or body_code == 110:
                return error.AuthenticationError()
            elif body_code == 111:
                return error.AccessTokenExpiredError()
            else:
                msg = "API error code {}: {}".format(resp["code"], resp["msg"])
                return error.APIError(msg)

        elif rcode == 503:
            raise error.ServiceUnavailableError(
                "The server is overloaded or not ready yet.",
                rbody,
                rcode,
                headers=rheaders,
            )

        elif rcode in [400, 404, 415]:
            return error.InvalidRequestError(
                error_msg,
                rbody,
                rcode,
                resp,
                rheaders,
            )

        else:
            return error.APIError(error_msg, rbody, rcode, resp, rheaders)

class ConsoleHTTPRequestor(HTTPRequestor):
    def __init__(self, ak=None, sk=None, request_type=None):
        self.ak = ak
        self.sk = sk
        self.version = wenxin_api.version
        self.request_type = request_type
        self.access_token_post_url = wenxin_api.variable.ACCESS_TOKEN_URL_CONSOLE

    def _get_access_token(self, ak=None, sk=None):
        if ak == None or sk == None:
            ak = wenxin_api.ak
            sk = wenxin_api.sk
        formatted_url = self.access_token_post_url + "?grant_type=client_credentials&client_id={}&client_secret={}".format(ak, sk)
        result = requests.get(formatted_url)
        if result.json().get("error", None):
            raise AuthenticationError(json.dumps(
                                        result.json(),
                                        indent=2))
        else:
            return result.json()["access_token"]

    def _resolve_response(self, rbody, rcode, rheaders) -> WenxinAPIResponse:

        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            data = json.loads(rbody)
        except (JSONDecodeError, UnicodeDecodeError):
            raise error.APIError(
                f"HTTP code {rcode} from API ({rbody})", rbody, rcode, headers=rheaders
            )

        if rcode != 200 or data["code"] != 0:
            raise self._request_error_handling(rbody, rcode, data, rheaders)

        if len(data["data"]) == 0:
            # some method ex. delete returns null data
            resp = WenxinAPIResponse(rheaders, self.request_type)
        elif self.request_type not in data["data"]:
            # some method  ex. create only returns ${type}_id
            resp = WenxinAPIResponse(rheaders, self.request_type, **data["data"])
        elif isinstance(data["data"][self.request_type], dict):
            resp = WenxinAPIResponse(rheaders, self.request_type, **data["data"][self.request_type])
        elif isinstance(data["data"][self.request_type], list):
            resp = [
                WenxinAPIResponse(rheaders, self.request_type, **one_data) \
                    for one_data in data["data"][self.request_type]
            ]
        else:
            raise ResponseDecodeError(json.dumps(data, ensure_ascii=False, indent=2))

        return resp

    def _request_error_handling(self, rbody, rcode, resp, rheaders):

        logger.error("wenxin api error {}: {}".format(rcode, rbody))
        error_msg = resp.get("error", "")
        return error.APIError(error_msg, rbody, rcode, resp, rheaders)



""" paragraph task """
from wenxin_api.api import Task
from wenxin_api.error import IllegalRequestArgumentError
from wenxin_api import log
logger = log.get_logger()

class Novel(Task):
    """ novel task """
    @staticmethod
    def _resolve_result(resp):
        rst = {}
        rst["result"] = resp["result"]
        return rst

    @classmethod
    def create(cls, text=None, model=None, **params):
        """ create """
        if text is None:
            raise IllegalRequestArgumentError("text shouldn't be none")
        model_id = 1 if model is None else model.id
        params["api_type"] = params.get("api_type", 6)
        logger.info("model {}: starts writing".format(model_id))
        resp = cls.default_request(text=text, **params)
        return cls._resolve_result(resp)